# -*- coding: utf-8 -*-
import re

from filetype import filetype

from pyramid_oereb.lib import b64


class ImageRecord(object):

    VALID_FILE_TYPES = [
        'png',
        'svg'
    ]  # type: list

    SVG_PATTERN = re.compile(r'<svg(.|\n)+<\/svg>')  # type: re

    """
    The record to hold the binary information of a image.

    Args:
        content (binary): The binary information of this image as binary string.
    """
    def __init__(self, content):

        self.content = content

    def encode(self):
        """
        Returns the image as base64 encoded string.

        Returns:
            str: The encoded image.
        """
        return b64.encode(self.content)

    @staticmethod
    def _validate_filetype(obj):
        """
        Checks for valid file type and returns extension and mime type.

        Args:
            obj (str, bytes, bytearray or readable): Path to the file, a readable object or the file's content
                as str, bytes or bytearray. See filetype documentation:
                https://h2non.github.io/filetype.py/v1.0.0/filetype.m.html

        Returns:
            tuple: The file's extension and mime type.

        Raises:
            TypeError: Raised if type is invalid or could not be recognized.
        """

        try:
            ft = filetype.guess(obj)
        except IOError:
            ft = None

        # Check for SVG content if filetype.guess fails
        if ft is None:

            # Test for readable object first
            try:
                content = obj.read()
            except AttributeError:
                # Test for file path
                try:
                    with open(obj) as f:
                        content = f.read()
                except Exception:
                    # Use input as content as last option
                    content = obj

            # Convert to string
            if isinstance(content, bytes) or isinstance(content, bytearray):
                content = content.decode('utf-8')

            # Check for SVG content
            if ImageRecord.SVG_PATTERN.search(content):
                result = ('svg', 'image/svg+xml')
            else:
                raise TypeError('Unrecognized file type')

        else:
            result = (ft.extension, ft.mime)

        # Check for extension in valid file types
        if result[0] not in ImageRecord.VALID_FILE_TYPES:
            raise TypeError('Invalid file type: {invalid}. Valid types are: {valid}.'.format(
                invalid=ft.extension,
                valid=ImageRecord.VALID_FILE_TYPES
            ))

        return result

    @staticmethod
    def get_extension(obj):
        """
        Checks for valid file type and returns its extension.

        Args:
            obj (str, bytes, bytearray or readable): Path to the file, a readable object or the file's content
                as str, bytes or bytearray. See filetype documentation:
                https://h2non.github.io/filetype.py/v1.0.0/filetype.m.html

        Returns:
            str: The file's extension.
        """
        return ImageRecord._validate_filetype(obj)[0]

    @staticmethod
    def get_mimetype(obj):
        """
        Checks for valid file type and returns its mime type.

        Args:
            obj (str, bytes, bytearray or readable): Path to the file, a readable object or the file's content
                as str, bytes or bytearray. See filetype documentation:
                https://h2non.github.io/filetype.py/v1.0.0/filetype.m.html

        Returns:
            str: The file's mime type.
        """
        return ImageRecord._validate_filetype(obj)[1]

    @property
    def mimetype(self):
        """
        Checks for valid file type and returns its extension.

        Returns:
            str: The file's extension.
        """
        return ImageRecord.get_mimetype(bytearray(self.content))

    @property
    def extension(self):
        """
        Checks for valid file type and returns its mime type.

        Returns:
            str: The file's mime type.
        """
        return ImageRecord.get_extension(bytearray(self.content))



# -*- coding: utf-8 -*-
import inspect
import optparse
import logging

from pyramid.config import ConfigurationError
from pyramid.path import DottedNameResolver
from sqlalchemy import create_engine

from pyramid_oereb.lib.config import Config
from pyramid_oereb.standard import create_tables_from_standard_configuration

logging.basicConfig()
log = logging.getLogger(__name__)


def _create_theme_tables(configuration_yaml_path, theme, section='pyramid_oereb', c2ctemplate_style=False,
                         tables_only=False):
    """
    Create all tables defined in the specified module.

    Args:
        configuration_yaml_path (str): Path to the configuration file.
        theme (str): Code of the theme to create the tables for.
        section (str): Section within the specified configuration file used for pyramid_oereb. Default is
            'pyramid_oereb'.
        c2ctemplate_style (bool): True if the yaml use a c2c template style (vars.[section]).
            Default is False.
        tables_only (bool): True to skip creation of schema. Default is False.
    """

    # Parse themes from configuration
    Config.init(configuration_yaml_path, section, c2ctemplate_style)
    themes = Config.get('plrs')
    if not isinstance(themes, list):
        raise ConfigurationError('No list of themes found.')

    # Find the specified theme
    found = False
    for t in themes:
        if t.get('code') == theme:

            # Check required configuration parameters
            params = t.get('source').get('params')
            if not isinstance(params, dict):
                raise ConfigurationError('Missing params property in source definition.')
            if not ('db_connection' in params and 'models' in params):
                raise ConfigurationError('Params has to contain "db_connection" and "models" properties.')

            # Create sqlalchemy engine for configured connection and load module containing models
            engine = create_engine(params.get('db_connection'), echo=True)
            models = DottedNameResolver().resolve(params.get('models'))

            if not tables_only:
                # Iterate over contained classes to collect needed schemas
                classes = inspect.getmembers(models, inspect.isclass)
                schemas = []
                create_schema_sql = 'CREATE SCHEMA IF NOT EXISTS {schema};'
                for c in classes:
                    class_ = c[1]
                    if hasattr(class_, '__table__') and class_.__table__.schema not in schemas:
                        schemas.append(class_.__table__.schema)

                # Try to create missing schemas
                connection = engine.connect()
                try:
                    for schema in schemas:
                        connection.execute(create_schema_sql.format(schema=schema))
                finally:
                    connection.close()

            # Create tables
            models.Base.metadata.create_all(engine)
            found = True
            break

    if not found:
        raise ValueError('Specified theme "{theme}" not found in configuration.'.format(theme=theme))


def create_standard_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all content for the standard database'
    )
    parser.add_option(
        '-c', '--configuration',
        dest='configuration',
        metavar='YAML',
        type='string',
        help='The absolute path to the configuration yaml file.'
    )
    parser.add_option(
        '-s', '--section',
        dest='section',
        metavar='SECTION',
        type='string',
        default='pyramid_oereb',
        help='The section which contains configuration (default is: pyramid_oereb).'
    )
    parser.add_option(
        '-T', '--tables-only',
        dest='tables_only',
        action='store_true',
        default=False,
        help='Use this flag to skip the creation of the schema.'
    )
    parser.add_option(
        '--sql-file',
        type='string',
        help='Generate an SQL file.'
    )
    parser.add_option(
        '--c2ctemplate-style',
        dest='c2ctemplate_style',
        action='store_true',
        default=False,
        help='Is the yaml file using a c2ctemplate style (starting with vars)'
    )
    options, args = parser.parse_args()
    if not options.configuration:
        parser.error('No configuration file set.')

    if options.sql_file is None:
        create_tables_from_standard_configuration(
            configuration_yaml_path=options.configuration,
            section=options.section,
            c2ctemplate_style=options.c2ctemplate_style,
            tables_only=options.tables_only
        )
    else:
        with open(options.sql_file, 'w') as sql_file:
            create_tables_from_standard_configuration(
                configuration_yaml_path=options.configuration,
                section=options.section,
                c2ctemplate_style=options.c2ctemplate_style,
                sql_file=sql_file
            )


def create_theme_tables():
    parser = optparse.OptionParser(
        usage='usage: %prog [options]',
        description='Create all tables for the specified theme.'
    )
    parser.add_option(
        '-c', '--configuration',
        dest='configuration',
        metavar='YAML',
        type='string',
        help='The absolute path to the configuration yaml file.'
    )
    parser.add_option(
        '-s', '--section',
        dest='section',
        metavar='SECTION',
        type='string',
        default='pyramid_oereb',
        help='The section which contains configuration (default is: pyramid_oereb).'
    )
    parser.add_option(
        '-t', '--theme',
        dest='theme',
        metavar='THEME_CODE',
        type='string',
        help='The theme code. Has to be available in configuration!'
    )
    parser.add_option(
        '-T', '--tables-only',
        dest='tables_only',
        action='store_true',
        default=False,
        help='Use this flag to skip the creation of the schema.'
    )
    parser.add_option(
        '--c2ctemplate-style',
        dest='tc2ctemplate_style',
        action='store_true',
        default=False,
        help='Is the yaml file using a c2ctemplate style (starting with vars)'
    )
    options, args = parser.parse_args()
    if not options.configuration:
        parser.error('No configuration file set.')
    if not options.theme:
        parser.error('No theme code defined.')
    try:
        _create_theme_tables(
            options.configuration,
            options.theme,
            section=options.section,
            c2ctemplate_style=options.c2ctemplate_style,
            tables_only=options.tables_only
        )
    except Exception as e:
        log.error(e)



# -*- coding: utf-8 -*-
import math
import numpy as np


def get_tag(element):
    return element.tag.split('}')[-1]


def parse_string(element, prop):
    for p in element:
        if get_tag(p) == prop:
            return p.text
    return None


def parse_multilingual_text(element, prop):
    for p in element:
        if get_tag(p) == prop:
            result = dict()
            for multilingual_text in p:
                for localization in multilingual_text:
                    for localized_text in localization:
                        language = text = None
                        for item in localized_text:
                            if get_tag(item) == 'Language':
                                language = item.text
                            else:
                                text = item.text
                        if language and text:
                            result[language] = text
            return result
    return None


def parse_ref(element, prop):
    for p in element:
        if get_tag(p) == prop:
            return p.attrib['REF']
    return None


def parse_article_numbers(document_reference, tag):
    for element in document_reference:
        if get_tag(element) == tag:
            numbers = list()
            for number in element:
                numbers.append(number.text)
            if len(numbers) > 0:
                return '|'.join(numbers)
    return None


def get_angle(d_start, d_arc, d_end):
    circle = 2 * math.pi
    d_start_reduced = 0.0
    d_arc_reduced = d_arc - d_start
    d_end_reduced = d_end - d_start
    if d_arc_reduced < 0:
        d_arc_reduced += circle
    elif d_arc_reduced >= circle:
        d_arc_reduced -= circle
    if d_end_reduced < 0:
        d_end_reduced += circle
    elif d_end_reduced >= circle:
        d_end_reduced -= circle
    if d_start_reduced <= d_arc_reduced <= d_end_reduced:
        return d_end_reduced - d_start_reduced, 1
    else:
        return circle - d_end_reduced - d_start_reduced, -1


def stroke_arc(start, arc, end, max_diff, precision):

    # Calculate shift (x, y) to reduce coordinate values onto average coordinate
    sum_x = start[0] + arc[0] + end[0]
    sum_y = start[1] + arc[1] + end[1]
    shift = (sum_x / 3.0, sum_y / 3.0)

    # Reduce initial coordinates
    start_reduced = (start[0] - shift[0], start[1] - shift[1])
    arc_reduced = (arc[0] - shift[0], arc[1] - shift[1])
    end_reduced = (end[0] - shift[0], end[1] - shift[1])

    # Calculate circle radius and center point
    les1 = np.array([
        [-1, start_reduced[0], start_reduced[1]],
        [-1, arc_reduced[0], arc_reduced[1]],
        [-1, end_reduced[0], end_reduced[1]]
    ])
    les2 = np.array([
        math.pow(start_reduced[0], 2) + math.pow(start_reduced[1], 2),
        math.pow(arc_reduced[0], 2) + math.pow(arc_reduced[1], 2),
        math.pow(end_reduced[0], 2) + math.pow(end_reduced[1], 2)
    ])
    a, b, c = np.linalg.solve(les1, les2)
    m = (b/2., c/2.)
    r = math.sqrt(math.pow(m[0], 2) + math.pow(m[1], 2) - a)

    # Calculate maximum angle step size
    max_delta_alpha = 2 * math.acos(1 - max_diff/r)

    # Calculate number of points needed to approximate the arc
    dx = end_reduced[0] - start_reduced[0]
    dy = end_reduced[1] - start_reduced[1]
    dist_start_end = math.sqrt(math.pow(dx, 2) + math.pow(dy, 2))
    diff_alpha = 2 * math.asin(dist_start_end / (2 * r))
    num_points = math.ceil(diff_alpha / max_delta_alpha)

    # Calculate directions
    d_start = math.atan2(start_reduced[1] - m[1], start_reduced[0] - m[0])
    d_arc = math.atan2(arc_reduced[1] - m[1], arc_reduced[0] - m[0])
    d_end = math.atan2(end_reduced[1] - m[1], end_reduced[0] - m[0])

    # Determine angle and rotation direction
    d_diff, rot = get_angle(d_start, d_arc, d_end)

    # Calculate points
    coords = list()
    d_step = d_diff / num_points
    for i in range(1, int(num_points) + 1, 1):
        x = m[0] + r * math.cos(d_start + d_step * i * rot) + shift[0]
        y = m[1] + r * math.sin(d_start + d_step * i * rot) + shift[1]
        coords.append((round(x, precision), round(y, precision)))

    # Remove duplicate points
    if len(coords) > 1:
        duplicates = list()
        for i in range(1, len(coords)):
            if coords[i][0] == coords[i - 1][0] and coords[i][1] == coords[i - 1][1]:
                duplicates.append(i)
        duplicates.reverse()
        for duplicate in duplicates:
            coords.pop(duplicate)

    # Return coordinates
    return coords
